
/* MZduino PC transfer software by Marq/Fit

1.10.2011: Different default port for Lin/Mac/Win, mzf support,
           send load/exec address, progress bar, slow down a bit
31.9.2011: Wait for the reset of the board, checksum calculation
29.9.2011: Initial version
*/

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/time.h>

#define BPS B57600      // Serial speed
#define SPEED 40        // us between bits
#define PROGRESS 1024   // Progress bar block size

// Default Arduino Uno port for Linux
#ifdef __linux__
#define PORT "/dev/ttyACM0"
#endif

// Mac OS X
#ifdef __APPLE__
#define PORT "/dev/cu.usbmodem621"
#endif

// The Microsoft attempt at an OS (untested)
#ifdef __WIN32__
#define PORT "COM3"
#endif

void put(int f,unsigned char c)
{
    struct timeval t;
    long long us1,us2;

    write(f,&c,1);
    tcdrain(f);

    gettimeofday(&t, NULL);
    us1=t.tv_sec*1000000+t.tv_usec;
    do
    {
        gettimeofday(&t, NULL);
        us2=t.tv_sec*1000000+t.tv_usec;
    }
    while(us2<us1+SPEED*5); // x4 for each bit and +1 for Arduino
}

// Spread data bits to two bytes with clock in every second bit
void encode(int f,unsigned char c)
{
    int n;
    unsigned char b=0x44;

    for(n=0;n<4;n++)
        if(c&(1<<n))
            b+=2<<n*2;
    put(f,b);

    b=0x44;
    for(n=0;n<4;n++)
        if(c&(0x10<<n))
            b+=2<<n*2;
    put(f,b);
}

int main(int argc,char *argv[])
{
    int f,i,c,
        size,jump,addr;
    unsigned char checksum=0,c2;
    struct termios opt;
    char *portname=PORT;
    FILE *s;

    printf("MZduino 0.04\n");

    if(argc<2)
    {
        printf("mzduino file.mzf [port]\n");
        return(1);
    }

    // Open the mzf
    s=fopen(argv[1],"rb");
    if(s==NULL)
    {
        printf("File can't be opened\n");
        return(1);
    }
    fseek(s,0,SEEK_END); // Find out data size (-128 for mzf header)
    size=ftell(s)-128;
    fseek(s,0,SEEK_SET);

    // Deal with the mzf header. Expect the file to be well-formed.
    fseek(s,20,SEEK_SET);
    addr=fgetc(s);      // Load address
    addr+=fgetc(s)<<8;
    jump=fgetc(s);      // Jump address
    jump+=fgetc(s)<<8;
    fseek(s,128,SEEK_SET); // Data starts here

    printf("Save parameters: %04x %04x %04x\n",addr,addr+size,jump);

    if(argc==3) // Override default port
        portname=argv[2];

    f=open(portname,O_RDWR|O_NOCTTY); // Open Arduino
    if(f==-1)
    {
        printf("Arduino can't be opened at %s\n",portname);
        return(1);
    }
    printf("Arduino opened\n");

    cfmakeraw(&opt); // Set serial params
    cfsetispeed(&opt,BPS);
    cfsetospeed(&opt,BPS);
    opt.c_cflag |= (CLOCAL | CREAD | CS8);
    opt.c_iflag |= ICRNL;

    if(tcsetattr(f, TCSANOW, &opt)<0)
    {
        printf("Serial attribs can't be set\n");
        return(1);
    }

    printf("Waiting for the board to reset\n");
    c2=' ';
    do
    {
        read(f,&c2,1);
    }
    while(c2!='!'); // The greeting message ends in '!'

    printf("Starting to transfer %d bytes",size);

    // "Handshake"
    put(f,0x00); // First some zero
    put(f,0x00);
    put(f,0xd0); // Then 1 and 3 (coded in two bits)

    // mzf parameters
    encode(f,size>>8);      // Data size
    encode(f,size&0xff);
    encode(f,jump>>8);      // Jump (start) address
    encode(f,jump&0xff);
    encode(f,addr>>8);      // Load address
    encode(f,addr&0xff);

    for(i=0;i<size;i++) // Byte by byte
    {
        c=fgetc(s);
        encode(f,c);
        checksum+=c;

        if((i&(PROGRESS-1))==0) // Progress bar
        {
            printf(".");
            fflush(stdout);
        }
    }

    encode(f,checksum);

    printf("\nDone! Checksum %x\n",checksum);
    close(f);
    fclose(s);
    return(0);
}
